def genHeader(domain):
    h  = '<!doctype html><html lang="en"><head>'
    h += '<title>Web Statistics and Analysis for ' + domain + '</title>'
    h += '<meta name="viewport" content="width=device-width, initial-scale=1">'
    h += '<meta charset="utf-8">'
    h += '<meta name="robots" content="index,follow">'
    h += '<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.3/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-QWTKZyjpPEjISv5WaRU9OFeRpok6YctnYmDr5pNlyT2bRjXh0JMhjY6hW+ALEwIH" crossorigin="anonymous">'
    h += '<script src="https://cdn.jsdelivr.net/npm/d3@7"></script>'
    h += '<script src="https://unpkg.com/vis-network/standalone/umd/vis-network.min.js"></script>'
    h += "\n" + '<script>'
    h += 'function DT2Locale(tdt) {'
    h += 'try {'
    h += 'var year = tdt.substring(0, 4);'
    # month is 0-indexed:
    h += 'var month = tdt.substring(4, 6) - 1;' 
    h += 'var day = tdt.substring(6, 8);'
    h += 'var hour = tdt.substring(8, 10);'
    h += 'var min = tdt.substring(10, 12);'
    h += 'var sec = tdt.substring(12, 14);'
    h += 'var tDateTime = new Date(year, month, day, hour, min, sec);'
    h += 'var tDT = tDateTime.toLocaleString();'
    h += 'tDT = tDT.replace(", 00:00:00","");'
    h += 'tDT = tDT.replace(", 12:00:00 AM","");'
    h += 'return tDT;'
    h += '} catch (error) {'
    h += ' return null; }'
    h += '}'
    h += 'function wsShowHide(eid) {'
    h += 'var x = document.getElementById(eid);'
    h += 'if (x.style.display == "none") { x.style.display = "block"; } else { x.style.display = "none"; }'
    h += '}'
    
    h += '</script>'
    h += '</head>' + "\n"
    h += '<body>'
    h += '<div class="container">'
    return h



import ast
from datetime import datetime

def parseRecJsonV0001(rec):
    r = ast.literal_eval(rec)    # parse rec to dict

    # local time:
    if 'time_local' in r:
        dto = datetime.strptime(r['time_local'], "%d/%b/%Y:%H:%M:%S %z")   # 07/Jan/2024:14:06:24 +0000
    elif 'tl' in r:
        dto = datetime.strptime(r['tl'], "%d/%b/%Y:%H:%M:%S %z")   # 07/Jan/2024:14:06:24 +0000
    else:
        raise Exception("no time found in input file. Details of input record: " + str(r) )

    ret = {
        'ymd': dto.strftime("%Y%m%d"),
        'timestamp': dto.strftime("%Y%m%d%H%M%S"),
        'referer': ''
    }

    # remote address IPv4 or IPv6:
    if 'ra' in r:
        ret['ip'] = r['ra']
    elif 'ip' in r:
        ret['ip'] = r['ip']
    elif 'remote_addr' in r:
        ret['ip'] = r['remote_addr']
    else:
        raise Exception("no remote addr found in input file.   Details of input record: " + str(r))

    # request:
    if 'rq' in r:
        ret['request'] = r['rq']
    elif 'request' in r:
        ret['request'] = r['request']
    elif 'u' in r:
        ret['request'] = r['u']
    elif 'ur' in r:
        ret['request'] = r['ur']
    else:
        raise Exception("no request found in input file.   for details check input record: " + str(r))
    
    if 'status' in r:
       ret['status'] = r['status']
    if 'st' in r:
       ret['status'] = r['st']

    if 'bytes_sent' in r:
        ret['bytes_sent'] = r['bytes_sent']
    if 'bs' in r:
        ret['bytes_sent'] = r['bs']
        
    if 'rf' in r:
         ret['referer'] =  r['rf']
    if 'http_referer' in r:
         ret['referer'] =  r['http_referer']
    
    if 'http_user_agent' in r: 
        ret['user_agent'] = r['http_user_agent']
    if 'ua' in r: 
        ret['user_agent'] = r['ua']
   
    # collect the optional attributes:
    if 'rt' in r and len(r['rt']) > 0:
        ret['request_time'] = r['rt']
    if 'request_time' in r and len(r['request_time']) > 0:
        ret['request_time'] = r['request_time']

    if 'upstream_respone_time' in r and len(r['upstream_response_time']) > 0:
        ret['response_time'] = r['upstream_respone_time']
    if 'rt' in r and len(r['rt']) > 0:
        ret['response_time'] = r['rt']
    
    if 'upstream_cache_status' in r:
        ret['cache_status'] = r['upstream_cache_status']
    if 'cs' in r:
        ret['cache_status'] = r['cs']
    
    if 'http_accept_language' in r:
        ret['accept_language'] = parse_accept_language(r['http_accept_language'])
    if 'al' in r:
        ret['accept_language'] = parse_accept_language(r['al'])
   
    return ret

#import re

# 'da, en-gb;q=0.8, en;q=0.7'  meaning: "I prefer Danish, but will accept British English and other types of English"
# 'fr-CH, fr;q=0.9, en;q=0.8, de;q=0.7, *;q=0.5'
# https://www.rfc-editor.org/rfc/rfc9110.html#name-accept-language
# https://www.rfc-editor.org/rfc/rfc4647.html
# https://www.rfc-editor.org/rfc/rfc5646.html
# Country code according to ISO3166-1 (capitalized letters)
# Language codes according to ISO639-1 (lowercase letters)

def parse_accept_language(accept_language_input):
    if not accept_language_input:
        return []

    if len(accept_language_input) > 8192:
        raise ValueError('Accept-Language exceeds maximum length of 8192 characters.')
        
    # Split the string by commas to get each language tag
    language_tags = accept_language_input.split(',')

    locales = {}
    #print("Input: " + accept_language_input)

    for tag in language_tags:
        # split locale and quality
        localeregion, _, qstr = tag.partition(';q=')
        if len(qstr) == 0:
            qstr = 1
            
        try:
            q = float(qstr or 1.0)
        except ValueError:
            continue  # ignore malformed entry

        # split locale into locale and region
        locale, _, region = localeregion.partition('-')
        
        #print("locale: " + str(locale) + " " + str(q) + " " + str(region) )
        if locale not in locales:
            locales[locale] = q
        
    #print("AL-Results: "  + str(locales))
    return locales



import argparse
import io
import json
import re
import sys
from datetime import datetime
from .updatestatistics import upd
from .updateStatisticsV0001 import updV0001
from .summarizemonth import summonth
from .summarizemonthV0001 import sumMonthV0001
from .migratev0001 import migv0001
from .migratev0001 import delv0000
from .parseRecJsonV0001 import parseRecJsonV0001
from .sumNavDayV0001 import sumNavDayV0001
from importlib.metadata import version
import geoip2.database

def parseRec(rec, log_pattern, j, georeader):
    r = {}
    # skip comments recs
    if rec[0:1] == '#':
        j['records_skipped_comment'] += 1
        return r,j

    if rec[0:2] == '{"':
        r = parseRecJsonV0001(rec)
    else:
        data = re.search(log_pattern, rec)

        if data:
            datadict = data.groupdict()
            ip_address = datadict["ipaddress"]
            timestamp = datadict["dateandtime"]
            request = datadict["url"]
            bytes_sent = datadict["bytessent"]
            referer = datadict["referer"]
            user_agent = datadict["useragent"]
            status = datadict["statuscode"]
            method = datadict["method"]
            if georeader:
                try:
                    grrsp = georeader.country(ip_address)
                    country = grrsp.country.name
                except geoip2.errors.AddressNotFoundError:
                    country = None
            else:
                country = None
        
            dto = datetime.strptime(timestamp,'%d/%b/%Y:%H:%M:%S %z')  # 07/Jan/2024:14:06:24 +0000
                
            r = {
                'ip': ip_address,
                'ymd': dto.strftime("%Y%m%d"),
                'timestamp': dto.strftime("%Y%m%d%H%M%S") ,
                'request': request,
                'status': status,
                'bytes_sent': 0,
                'referer': referer,
                'user_agent': user_agent
            }
            if country:
                r['country'] = country
            if bytes_sent.isdigit():                 # bytes_sent can also hold "-"
                r['bytes_sent'] = int(bytes_sent)
        else:
            print("Log Rec parsing failed for: " + rec[0:79])
            j['records_parsing_failed'] += 1
        
    return r,j


def runws(statfile, infile, geoip, verbosity, domain):

    try: 
        georeader = geoip2.database.Reader(geoip)
    except (FileNotFoundError, TypeError) as gerr:
        georeader = None
        print("geoip2 file not found, continue processing")

    # init statistic file if it does not exist:
    d = {}
    d['timelastrec'] = '19991231235959'
    d['v0001'] = {}
    d['v0001']['days'] = {}

    # init job results
    j = {
        'records_read': 0,
        'records_skipped_comment': 0,
        'records_parsing_failed': 0,
        'records_already_processed': 0,
        'records_processed_for_statistic': 0,
    }
    
    # load statistic file if it exists
    try:
        f = open(statfile) 
        d = json.load(f) 
    except FileNotFoundError:   # first call: file does not exists
        print("-s statistic file not found, it will be automatically created")
    except json.JSONDecodeError:
        print("-s json file is not valid")

    visitIP = {}
    lasttimerecobj = datetime.strptime(d['timelastrec'],"%Y%m%d%H%M%S")
    print("lasttimerecobj: " + str(lasttimerecobj))

    log_pattern = re.compile(
        r"""
        (?P<ipaddress>
        \d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|               # ipv4
        (                                                  
        ([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|          # ipv6 1:2:3:4:5:6:7:8
        ([0-9a-fA-F]{1,4}:){1,7}:|                         # ipv6 1::              1:2:3:4:5:6:7::
        ([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|         # ipv6 1::8             1:2:3:4:5:6::8  1:2:3:4:5:6::8
        ([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|  # ipv6 1::7:8           1:2:3:4:5::7:8  1:2:3:4:5::8
        ([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|  # ipv6 1::6:7:8         1:2:3:4::6:7:8  1:2:3:4::8
        ([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|  # ipv6 1::5:6:7:8       1:2:3::5:6:7:8  1:2:3::8
        ([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|  # ipv6 1::4:5:6:7:8     1:2::4:5:6:7:8  1:2::8
        [0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|       # ipv6 1::3:4:5:6:7:8   1::3:4:5:6:7:8  1::8  
        :((:[0-9a-fA-F]{1,4}){1,7}|:)|                     # ipv6 ::2:3:4:5:6:7:8  ::2:3:4:5:6:7:8 ::8       ::     
        fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|     # fe80::7:8%eth0   fe80::7:8%1  (link-local IPv6 addresses with zone index)
        ::(ffff(:0{1,4}){0,1}:){0,1}
        ((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}
        (25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|          # ::255.255.255.255   ::ffff:255.255.255.255  ::ffff:0:255.255.255.255  (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
        ([0-9a-fA-F]{1,4}:){1,4}:
        ((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}
        (25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])           # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33 (IPv4-Embedded IPv6 Address)
        )
        )[ ]-[ ](?P<uid>.+)[ ]\[(?P<dateandtime>\d{2}\/[a-zA-Z]{3}\/\d{4}:\d{2}:\d{2}:\d{2}[ ](\+|\-)\d{4})\][ ]
        ((?P<method>)(\"(GET|POST|HEAD|PUT|DELETE|OPTIONS|PROPFIND))[ ]
        (?P<url>.+)[ ](HTTP\/(1\.0|1\.1|2\.0)"))[ ]
        (?P<statuscode>\d{3})[ ]
        (?P<bytessent>\S+)[ ]                              # number, can also be "-"
        (["](?P<referer>[^"]+)["])[ ]
        (["](?P<useragent>[^"]+)["])
        """,
        flags=re.VERBOSE
    )
    
    # process infile:
    with open(infile,'r') as infile:
        for rec in infile:
            j['records_read'] += 1
            recparsed, j = parseRec(rec, log_pattern, j, georeader)
            # skip unrecognized records:
            if not recparsed or recparsed['timestamp'] is None or recparsed['ip'] is None:
                continue
            # skip already processed data:
            if recparsed['timestamp']  <=  d['timelastrec']:
                j['records_already_processed'] += 1
                continue
            d, visitIP = updV0001(d, recparsed, visitIP, domain)
            j['records_processed_for_statistic'] += 1
            

    # sum navigation
    d = sumNavDayV0001(d)
    
    # summarize previous months
    d = summonth(d)
    d = sumMonthV0001(d, statfile)
    
    
    # write updated statistic file:
    with open(statfile, "w") as sf:
       json.dump(d,sf)  

    print("Rec read:           "      + str(j['records_read']))
    print("Rec processed:      " + str(j['records_processed_for_statistic']))
    print("Rec parsing failed: " + str(j['records_parsing_failed']))
    print("Rec skipped (already processed): "   + str(j['records_already_processed']))
    return 0

if __name__ == "__main__":
    parser = argparse.ArgumentParser(allow_abbrev=False,
        prog='a9x_webstatistics',
        epilog="Version: "+ version('a9x-webstatistics')
    )
    parser.add_argument("-s", "--statfile", help="json file that contains calculated statistics", default="webstat.json")
    parser.add_argument("-i", "--infile", help="filename including path to web server access log that contains input data", default="nginx_access.log")
    parser.add_argument("-g", "--geoip", help="path/file to GeoIP2-Country.mmdb file")
    parser.add_argument("-v", "--verbose", help="increase output verbosity, 0=none, 1=increased verbosity", default="0")
    parser.add_argument("-d", "--domain", help="domain https://logikfabrik.com on which the access log file runs", default="https://logikfabrik.com")
    args, unknown = parser.parse_known_args()

    migv0001(statfile=args.statfile)
    runws(statfile=args.statfile, infile=args.infile, geoip=args.geoip, verbosity=args.verbose, domain=args.domain)
    delv0000(statfile=args.statfile)



# -*- coding: utf-8 -*-
from operator import itemgetter

# navigation chart:
def navchart(d, owndomain, omit):
    h = ''
    links = []    # {'s': 'google.com', 't': '/team/view/ax', 'c': 1}
    nodes = []    # {'id': 'google.com', 'y':'root'}
    days = 0
    for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        if len(k) < 8:   # not a day anymore
            break
        if 'user' in d['v0001']['days'][k] and 'navigation' in d['v0001']['days'][k]['user']:
            for sk,sv in d['v0001']['days'][k]['user']['navigation'].items():
                n = sk.split('(())') 
                if n[0] == n[1]:
                    continue
                # allow only a-z and 0-9:
                if n[0] == '/':    # to avoid empty na
                    n[0] = owndomain
                if n[1] == '/':    # to avoid empty nb
                    n[1] = owndomain
                tmplink = {}
                tmplink['s'] = n[0]
                tmplink['t'] = n[1]
                tmplink['c'] = sv
                duplicate_found = False
                for li in links:
                    if (li['s'] == tmplink['s']
                            and li['t'] == tmplink['t']):
                        duplicate_found = True
                        li['c'] += sv
                        break
                if duplicate_found == False:
                    links.append(tmplink)
                   
                tmpnode = {}
                tmpnode['id'] = n[0]
                tmpnode['y'] = 'root'
                duplicate_found = False
                for no in nodes:
                    if (no['id'] == tmpnode['id']):
                        duplicate_found = True
                        break
                if duplicate_found == False:
                    nodes.append(tmpnode)
            
        days += 1
        if days > 31:
            break
 
    # d3js horizontal bubble char in case results are available:
    #if len(toplng) > 0:
    if 'logikfabrik' in owndomain:
        h += "\n\n"
        h += '<div class="col-md-12 col-lg-12 col-xxl-12">'
        h += '<div class="card mt-2"><div class="card-body">'
        h += '<h3 class="card-title">Navigation Chart</h3>'
        h += '<p class="card-text">User Navigation Chart for ' + owndomain + ':'
        h += '<div id="navchart-container"><svg id="svgchart" width="600" height="400"></svg></div>'
        h += '<script type="module">'
        h += 'const nodes = ' + str(nodes) + ';'
        h += 'const links = ' + str(links) + ';'
        h += 'const rect = document.getElementById("navchart-container").getBoundingClientRect();'
        h += 'const margin = { top: 20, right: 20, bottom: 40, left: 100 };'
        h += 'const width = Math.round(rect.width) - margin.left - margin.right;'
        h += 'const height = 400 - margin.top - margin.bottom;'

        # 2. Set up the SVG container and dimensions
        h += 'const svg = d3.select("#svgchart");'
        h += "const width = +svg.attr('width');"
        h += "const height = +svg.attr('height');"
       
        # 3. Set up the D3 force simulation
        h += "const simulation = d3.forceSimulation(nodes)"
        h += ".force('link', d3.forceLink(links).id(d => d.id).distance(150))"
        h += ".force('charge', d3.forceManyBody().strength(-300))"
        h += ".force('center', d3.forceCenter(width / 2, height / 2));"

        # 4. Create links (edges)
        h += "const link = svg.append('g')"
        h += ".selectAll('.link')"
        h += ".data(links)"
        h += ".enter().append('line')"
        h += ".attr('class', 'link');"

        # 6. Add labels to the nodes
        h += "const labels = svg.append('g')"
        h += ".selectAll('.label')"
        h += ".data(nodes)"
        h += ".enter().append('text')"
        h += ".attr('class', 'label')"
        h += ".attr('dx', 25)"
        h += ".attr('dy', '.35em')"
        h += ".text(d => d.name);"

        # 7. Place root nodes along the left side of the screen and freeze their positions
        h += "const rootNodes = nodes.filter(n => n.typ === 'root');"
        h += "rootNodes.forEach((node, index) => {"
        h += "node.x = 100; // Set all root nodes on the left (x = 100)"
        h += "node.y = 100 + index * 100;"   # Distribute vertically along a line
        h += "node.fx = node.x;"  # Freeze position
        h += "node.fy = node.y;"  # Freeze position
        h += "});"

        # 8. Define the tick function to update positions
        h += "simulation.on('tick', function() {"
        # Update link positions
        h += "link"
        h += ".attr('x1', d => d.source.x)"
        h += ".attr('y1', d => d.source.y)"
        h += ".attr('x2', d => d.target.x)"
        h += ".attr('y2', d => d.target.y);"

        # Update node positions
        h += "node.attr('cx', d => d.x).attr('cy', d => d.y);"

        # Update label positions
        h += "labels.attr('x', d => d.x).attr('y', d => d.y);"
        h += "});"

        # 9. Drag functions to allow node movement
        h += "function dragStarted(event) {"
        h += "if (!event.active) simulation.alphaTarget(0.3).restart();"
        h += "event.subject.fx = event.subject.x;"
        h += "event.subject.fy = event.subject.y;"
        h += "}"

        h += "function dragged(event) {"
        h += "event.subject.fx = event.x;"
        h += "event.subject.fy = event.y;"
        h += "}"

        h += "function dragEnded(event) {"
        h += "if (!event.active) simulation.alphaTarget(0);"
        h += "event.subject.fx = null;"
        h += "event.subject.fy = null;"
        h += "}"
        h += "</script>"

        h += '</p>'
        h += '</div></div></div>'   # end of card and col
    return h
